"""ble.connection.py

Copyright (c) 2019, Douglas Otwell, DashIO-Connect

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import json
import logging
import threading
import time
import uuid
import shortuuid

import dbus
import dbus.exceptions
import dbus.mainloop.glib
import dbus.service
import zmq
from gi.repository import GLib

from dashio.device import Device

from .constants import CONNECTION_PUB_URL


class BLEControl():
    """BLE connection control
    """
    def get_state(self) -> str:
        """A CFG only control"""
        return ""

    def get_cfg(self, data) -> str:
        """Called by iotdashboard app to get controls CFG
        """
        try:
            dashboard_id = data[2]
        except IndexError:
            return ""
        cfg_str = f"\tCFG\t{dashboard_id}\t" + self.cntrl_type + "\t" + json.dumps(self._cfg) + "\n"
        return cfg_str

    def get_cfg64(self, data) -> dict:
        """Returns the CFG dict for this TCP control

        Returns
        -------
        dict
            The CFG string for this control
        """
        return self._cfg

    def __init__(self, control_id, service_uuid="", read_uuid="", write_uuid=""):
        self._cfg = {}
        self.cntrl_type = "BLE"
        self._cfg["controlID"] = control_id
        self.control_id = control_id
        self.service_uuid = service_uuid
        self.read_uuid = read_uuid
        self.write_uuid = write_uuid

    @property
    def service_uuid(self) -> str:
        """ServiceUUID"""
        return self._cfg["serviceUUID"]

    @service_uuid.setter
    def service_uuid(self, val: str):
        self._cfg["serviceUUID"] = val

    @property
    def read_uuid(self) -> str:
        """readUUID"""
        return self._cfg["readUUID"]

    @read_uuid.setter
    def read_uuid(self, val: str):
        self._cfg["readUUID"] = val

    @property
    def write_uuid(self) -> str:
        """writeUUID"""
        return self._cfg["writeUUID"]

    @write_uuid.setter
    def write_uuid(self, val: str):
        self._cfg["writeUUID"] = val


GATT_CHRC_IFACE = "org.bluez.GattCharacteristic1"
NOTIFY_TIMEOUT = 10
BLUEZ_SERVICE_NAME = "org.bluez"
LE_ADVERTISING_MANAGER_IFACE = "org.bluez.LEAdvertisingManager1"
DBUS_OM_IFACE = "org.freedesktop.DBus.ObjectManager"
DBUS_PROP_IFACE = "org.freedesktop.DBus.Properties"
LE_ADVERTISEMENT_IFACE = "org.bluez.LEAdvertisement1"
GATT_MANAGER_IFACE = "org.bluez.GattManager1"
GATT_SERVICE_IFACE = "org.bluez.GattService1"
GATT_DESC_IFACE = "org.bluez.GattDescriptor1"


class DashIOAdvertisement(dbus.service.Object):
    """BLE Advertisement
    """
    PATH_BASE = "/org/bluez/example/advertisement"

    def __init__(self, index, service_uuid):
        self.path = self.PATH_BASE + str(index)
        self.bus = dbus.SystemBus()
        self.service_uuids = []
        self.service_uuids.append(service_uuid)
        self.properties = {}
        self.properties["Type"] = "peripheral"
        self.properties["LocalName"] = dbus.String("DashIO")
        self.properties["ServiceUUIDs"] = dbus.Array(self.service_uuids, signature='s')
        self.properties["IncludeTxPower"] = dbus.Boolean(True)
        dbus.service.Object.__init__(self, self.bus, self.path)
        self.register()

    def find_adapter(self, bus):
        """Find BLE adapters"""
        remote_om = dbus.Interface(bus.get_object(BLUEZ_SERVICE_NAME, "/"), DBUS_OM_IFACE)
        objects = remote_om.GetManagedObjects()

        for obj, props in objects.items():
            if LE_ADVERTISING_MANAGER_IFACE in props:
                return obj
        return None

    def get_properties(self):
        """Return the properties dictionary"""
        return {LE_ADVERTISEMENT_IFACE: self.properties}

    def get_path(self):
        """Get the DBUS Path to BLE advertisement"""
        return dbus.ObjectPath(self.path)

    # pylint: disable=invalid-name
    @dbus.service.method(DBUS_PROP_IFACE, in_signature="s", out_signature="a{sv}")
    def GetAll(self, iface):
        """DBUS calls this to retrieve Advertisements"""
        if iface != LE_ADVERTISEMENT_IFACE:
            raise InvalidArgsException()
        return self.get_properties()[LE_ADVERTISEMENT_IFACE]

    # pylint: disable=invalid-name
    @dbus.service.method(LE_ADVERTISEMENT_IFACE, in_signature='', out_signature='')
    def Release(self):
        """DBUS calls this on release"""
        logging.debug('%s: Released!', self.path)

    def register_ad_callback(self):
        """DBUS calls this on release"""
        logging.debug("GATT advertisement registered")

    def register_ad_error_callback(self):
        """DBUS calls this on error"""
        logging.debug("Failed to register GATT advertisement")

    def register(self):
        """Register with DBUS"""
        bus = self.bus
        adapter = self.find_adapter(bus)
        ad_manager = dbus.Interface(bus.get_object(BLUEZ_SERVICE_NAME, adapter), LE_ADVERTISING_MANAGER_IFACE)
        ad_manager.RegisterAdvertisement(self.get_path(), {}, reply_handler=self.register_ad_callback, error_handler=self.register_ad_error_callback)


class InvalidArgsException(dbus.exceptions.DBusException):
    """DBUS errors"""
    _dbus_error_name = "org.freedesktop.DBus.Error.InvalidArgs"


class NotSupportedException(dbus.exceptions.DBusException):
    """DBUS errors"""
    _dbus_error_name = "org.bluez.Error.NotSupported"


class NotPermittedException(dbus.exceptions.DBusException):
    """DBUS errors"""
    _dbus_error_name = "org.bluez.Error.NotPermitted"


class BLEConnection(dbus.service.Object, threading.Thread):
    """BLEConnection
    """
    def _send_announce(self):
        msg = {
            'msgType': 'send_announce',
            'connectionUUID': self.zmq_connection_uuid
        }
        #  logging.debug("TCP SEND ANNOUNCE: %s", msg)
        self.tx_zmq_pub.send_multipart([b"COMMAND", json.dumps(msg).encode()])

    def add_device(self, device: Device):
        """Add a device to the connection

        Parameters
        ----------
        device : dashio.Device
            Add a device to the connection.
        """
        if device.device_id not in self.local_device_id_list:
            device.register_connection(self)
            self.local_device_id_list.append(device.device_id)

    def close(self):
        """Close the connection
        """
        logging.debug("\nGATT application terminated")
        self.tx_zmq_pub.close()
        self.rx_zmq_sub.close()
        self.mainloop.quit()

    def _zmq_callback(self, queue, condition):
        # logging.debug('zmq_callback')

        while self.rx_zmq_sub.getsockopt(zmq.EVENTS) & zmq.POLLIN:
            try:
                [msg_to, data] = self.rx_zmq_sub.recv_multipart()
            except ValueError:
                continue
            if not data:
                continue

            data_str = data.decode('utf-8')
            # TODO: need to set this for the negitiated MTU
            # 160 seems to work with iPhone
            mtu = 160
            data_chunks = [data_str[i: i + mtu] for i in range(0, len(data_str), mtu)]

            sent = False
            for data_chunk in data_chunks:
                sent |= self.dash_service.dash_characteristics.ble_send(data_chunk)
            if sent:
                logging.debug("BLE TX: %s", data_str.strip())

        return True

    def _ble_rx(self, msg: str):
        msg_from = self.b_connection_id
        #  logging.debug("BLE ZMQ TX: %s", msg)
        self.tx_zmq_pub.send_multipart([msg.encode(), msg_from])

    def __init__(self, ble_uuid=None, context: zmq.Context = None):
        """BLE Connection

        Parameters
        ----------
        ble_uuid : str, optional
            The UUID used by the BLE connection, if None a UUID is generated
        context : ZMQ Context, optional
            ZMQ Context, by default None
        """
        threading.Thread.__init__(self, daemon=True)

        self.zmq_connection_uuid = "BLE:" + shortuuid.uuid()
        self.b_connection_id = self.zmq_connection_uuid.encode('utf-8')
        self.local_device_id_list = []
        self.context = context or zmq.Context.instance()
        self.rx_zmq_sub = self.context.socket(zmq.SUB)
        self.rx_zmq_sub.setsockopt_string(zmq.SUBSCRIBE, "ALL")
        self.rx_zmq_sub.setsockopt_string(zmq.SUBSCRIBE, "COMMAND")
        self.rx_zmq_sub.setsockopt_string(zmq.SUBSCRIBE, "BLE")
        self.rx_zmq_sub.setsockopt_string(zmq.SUBSCRIBE, self.zmq_connection_uuid)
        # TODO: Need to figure out why this doesn't work
        # GLib.io_add_watch(
        #     self.rx_zmq_sub.getsockopt(zmq.FD),
        #     GLib.IO_IN | GLib.IO_ERR | GLib.IO_HUP | GLib.IO_PRI,
        #     self.zmq_callback
        # )
        GLib.timeout_add(10, self._zmq_callback, "q", "p")
        self.tx_zmq_pub = self.context.socket(zmq.PUB)
        self.tx_zmq_pub.bind(CONNECTION_PUB_URL.format(id=self.zmq_connection_uuid))
        dashio_service_uuid = ble_uuid or str(uuid.uuid4())

        # self.connection_control = BLEControl(self.zmq_connection_uuid, dashio_service_uuid, str(uuid.uuid4()), str(uuid.uuid4()))

        GLib.threads_init()
        dbus.mainloop.glib.threads_init()
        dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
        self.mainloop = GLib.MainLoop()

        self.bus = dbus.SystemBus()
        self.path = "/"
        self.dash_service = DashIOService(0, dashio_service_uuid, self._ble_rx)
        self.response = {}

        chrc = self.dash_service.get_characteristics()
        self.response[chrc.get_path()] = chrc.get_properties()
        self.response[self.dash_service.get_path()] = self.dash_service.get_properties()

        dbus.service.Object.__init__(self, self.bus, self.path)
        self._register()
        self.adv = DashIOAdvertisement(0, dashio_service_uuid)
        self.start()
        time.sleep(0.5)

    def _find_adapter(self, bus):
        remote_om = dbus.Interface(bus.get_object(BLUEZ_SERVICE_NAME, "/"), DBUS_OM_IFACE)
        objects = remote_om.GetManagedObjects()

        for obj, props in objects.items():
            if LE_ADVERTISING_MANAGER_IFACE in props:
                return obj
        return None

    def get_path(self):
        """Get DBUS path"""
        return dbus.ObjectPath(self.path)

    # pylint: disable=invalid-name
    @dbus.service.method(DBUS_OM_IFACE, out_signature="a{oa{sa{sv}}}")
    def GetManagedObjects(self):
        """Called by DBUS"""
        return self.response

    def _register_app_callback(self):
        logging.debug("GATT application registered")

    def _register_app_error_callback(self, error):
        logging.debug("Failed to register application: %s", str(error))

    def _register(self):
        adapter = self._find_adapter(self.bus)
        service_manager = dbus.Interface(self.bus.get_object(BLUEZ_SERVICE_NAME, adapter), GATT_MANAGER_IFACE)
        service_manager.RegisterApplication(self.get_path(), {}, reply_handler=self._register_app_callback, error_handler=self._register_app_error_callback)

    def run(self):
        self._send_announce()
        self.mainloop.run()


class DashIOService(dbus.service.Object):
    """DBUS DashIO BLE service
    """
    PATH_BASE = "/org/bluez/example/service"

    def __init__(self, index, service_uuid, ble_rx):
        self.bus = dbus.SystemBus()
        self.path = self.PATH_BASE + str(index)
        self.uuid = service_uuid
        self.primary = True

        self.dash_characteristics = DashConCharacteristic(self, service_uuid, ble_rx)

        dbus.service.Object.__init__(self, self.bus, self.path)

    def get_properties(self):
        """Get DBUS properties"""
        return {
            GATT_SERVICE_IFACE: {
                'UUID': self.uuid,
                'Primary': self.primary,
                'Characteristics': dbus.Array(
                    self.get_characteristic_paths(),
                    signature='o')
            }
        }

    def get_path(self):
        """Get DBUS path"""
        return dbus.ObjectPath(self.path)

    def get_characteristic_paths(self):
        """get DBUS properties"""
        result = []
        result.append(self.dash_characteristics.get_path())
        return result

    def get_characteristics(self):
        """Get Dashio charactoristics"""
        return self.dash_characteristics

    def get_bus(self):
        """Return the DBUS bus"""
        return self.bus

    # pylint: disable=invalid-name
    @dbus.service.method(DBUS_PROP_IFACE, in_signature='s', out_signature='a{sv}')
    def GetAll(self, iface):
        """Called by DBUS"""
        if iface != GATT_SERVICE_IFACE:
            raise InvalidArgsException()
        return self.get_properties()[GATT_SERVICE_IFACE]


class DashConCharacteristic(dbus.service.Object):
    """
    org.bluez.GattCharacteristic1 interface implementation
    """
    def __init__(self, service, chacteristic_uuid, ble_rx):
        self.path = service.path + '/char' + str(1)
        self.bus = service.get_bus()
        self.uuid = chacteristic_uuid
        self.service = service
        self.flags = ["notify", "write-without-response"]
        self.notifying = False
        self._ble_rx = ble_rx
        self.read_buffer = ""
        dbus.service.Object.__init__(self, self.bus, self.path)

    def get_properties(self):
        """Get DBUS properties"""
        return {
            GATT_CHRC_IFACE: {
                'Service': self.service.get_path(),
                'UUID': self.uuid,
                'Flags': self.flags
            }
        }

    def get_path(self):
        """Get DBUS path"""
        return dbus.ObjectPath(self.path)

    # pylint: disable=invalid-name
    @dbus.service.method(DBUS_PROP_IFACE, in_signature='s', out_signature='a{sv}')
    def GetAll(self, iface):
        """Called by DBUS"""
        if iface != GATT_CHRC_IFACE:
            raise InvalidArgsException()
        return self.get_properties()[GATT_CHRC_IFACE]

    # pylint: disable=invalid-name
    @dbus.service.method(GATT_CHRC_IFACE, in_signature='a{sv}', out_signature='ay')
    def ReadValue(self, _):
        """Called by DBUS"""
        logging.debug('Default ReadValue called, returning error')
        raise NotSupportedException()

    # pylint: disable=invalid-name, unnecessary-pass
    @dbus.service.signal(DBUS_PROP_IFACE, signature='sa{sv}as')
    def PropertiesChanged(self, interface, changed, invalidated):
        """Called by DBUS"""
        pass

    def get_bus(self):
        """Return DBUS bus"""
        bus = self.bus
        return bus

    def ble_send(self, tx_data):
        """Send data to BLE connection"""
        if self.notifying:
            value = [dbus.Byte(c.encode()) for c in tx_data]
            self.PropertiesChanged(GATT_CHRC_IFACE, {"Value": value}, [])
        return self.notifying

    # pylint: disable=invalid-name
    @dbus.service.method(GATT_CHRC_IFACE)
    def StartNotify(self):
        """Called by DBUS"""
        if self.notifying:
            return
        self.notifying = True

    # pylint: disable=invalid-name
    @dbus.service.method(GATT_CHRC_IFACE)
    def StopNotify(self):
        """Called by DBUS"""
        self.notifying = False

    # pylint: disable=invalid-name
    @dbus.service.method(GATT_CHRC_IFACE, in_signature='aya{sv}')
    def WriteValue(self, value, options):
        """Called by DBUS"""
        rx_str = ''.join([str(v) for v in value])
        self.read_buffer += rx_str
        if rx_str[-1] == '\n':
            logging.debug("BLE RX: %s", self.read_buffer.strip())
            self._ble_rx(self.read_buffer)
            self.read_buffer = ''



import json
import unittest

from dashio import Device


class TestDashDevice(unittest.TestCase):
    def _get_cfg_dict(self, cfg_str):
        json_str = cfg_str.rpartition('\t')[2]
        return json.loads(json_str)

    def test_dash_device_type(self):
        test_device = Device("DEVICETYPE", "DEVICEID", "DEVICENAME")
        self.assertEqual(test_device.device_type, 'DEVICETYPE', "device_type type should be DEVICETYPE")

    def test_dash_device_id(self):
        test_device = Device("DEVICETYPE", "DEVICEID", "DEVICENAME")
        self.assertEqual(test_device.device_id, 'DEVICEID', "device_id type should be DEVICEID")

    def test_dash_device_name(self):
        test_device = Device("DEVICETYPE", "DEVICEID", "DEVICENAME")
        self.assertEqual(test_device._device_name, 'DEVICENAME', "_device_name type should be DEVICENAME")

    def test_dash_device_cfg_number_of_controls_boxes(self):
        test_device = Device("DEVICETYPE", "DEVICEID", "DEVICENAME",)
        self.assertEqual(test_device._cfg['numDeviceViews'], 0, "editLock type should be 0")


if __name__ == '__main__':
    unittest.main()



import json
import unittest


class TestTCPConnection(unittest.TestCase):
    def _get_cfg_dict(self, cfg_list: list):
        json_str = cfg_list[0].rpartition('\t')[2]
        return json.loads(json_str)


if __name__ == '__main__':
    unittest.main()



"""
MIT License

Copyright (c) 2020 DashIO-Connect

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import json
import logging
import socket
import threading
import time

import shortuuid
import zmq

from . import ip
from .constants import CONNECTION_PUB_URL
from .device import Device
from .zeroconf_service import ZeroconfService


class TCPConnection(threading.Thread):
    """Setups and manages a connection thread to iotdashboard via TCP."""

    def _send_announce(self):
        msg = {
            'msgType': 'send_announce',
            'connectionUUID': self.zmq_connection_uuid
        }
        # logging.debug("TCP SEND ANNOUNCE: %s", msg)
        self.tx_zmq_pub.send_multipart([b"COMMAND", json.dumps(msg).encode()])

    def add_device(self, device: Device):
        """Add a device to the connection

        Parameters
        ----------
        device : dashio.Device
            Add a device to the connection.
        """
        if device.device_id not in self.local_device_id_list:
            device.register_connection(self)
            self.local_device_id_list.append(device.device_id)
            self.z_conf.add_device(device.device_id)

    @staticmethod
    def _is_port_in_use(ip_address, port):
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as port_s:
            return port_s.connect_ex((ip_address, port)) == 0

    def __init__(self, ip_address="*", port=5650, use_zero_conf=True, context: zmq.Context = None):
        """TCP Connection

        Parameters
        ---------
            ip_address : str, optional
                IP Address to use. Defaults to "*" - forces The TCP connection to find the ip address attached to the local network.
            port : int, optional
                Port to use. Defaults to 5650.
            use_zero_conf : bool, optional
                Use mDNS to advertise the connection. Defaults to True.
            context : optional
                ZMQ context. Defaults to None.
        """

        threading.Thread.__init__(self, daemon=True)
        self.context = context or zmq.Context.instance()
        self.zmq_connection_uuid = "TCP:" + shortuuid.uuid()
        self.b_zmq_connection_uuid = self.zmq_connection_uuid.encode('utf-8')
        self.use_zeroconf = use_zero_conf

        if ip_address == "*":
            self.local_ip = ip.get_local_ip_address()
        else:
            self.local_ip = ip_address
        self.local_port = port
        while self._is_port_in_use(self.local_ip, self.local_port) and use_zero_conf:
            # increment port until we find one that is free.
            self.local_port += 1
        self.ext_url = "tcp://" + self.local_ip + ":" + str(self.local_port)

        self.socket_ids = []
        self.local_device_id_list = []
        self.remote_connection_dict = {}
        self.remote_device_con_dict = {}
        self.remote_device_id_msg_dict = {}
        self._device_id_action_station_list = []

        self.tcp_id_2_ip_dict = {}
        self.tcp_ip_2_id_dict = {}

        self.running = True

        self.tx_zmq_pub = self.context.socket(zmq.PUB)
        self.tx_zmq_pub.bind(CONNECTION_PUB_URL.format(id=self.zmq_connection_uuid))

        if self.use_zeroconf:
            self.z_conf = ZeroconfService(self.zmq_connection_uuid, self.local_ip, self.local_port, self.context)

        self.start()

    def close(self):
        """Close the connection."""

        if self.use_zeroconf:
            self.z_conf.close()
        self.running = False

    def _connect_remote_device(self, msg: dict):
        """Connect to remote device"""
        ip_address = msg['address']
        port = msg['port']
        url = f"tcp://{ip_address}:{port}"
        self.tcpsocket.connect(url)
        socket_id = self.tcpsocket.getsockopt(zmq.IDENTITY)
        ip_b = ip_address + ':' + port
        if ip_b not in self.tcp_ip_2_id_dict:
            try:
                logging.debug("Sending WHO to ID: %s", socket_id.hex())
                self.tcpsocket.send(socket_id, zmq.SNDMORE)
                self.tcpsocket.send_string('\tWHO\n')
                self.tcp_id_2_ip_dict[socket_id] = ip_address + ':' + port
                self.tcp_ip_2_id_dict[ip_address + ':' + port] = socket_id
            except zmq.error.ZMQError:
                logging.debug("Sending TX Error.")
                self.tcpsocket.send(b'')
        time.sleep(0.1)
        if socket_id not in self.socket_ids:
            logging.debug("Added Socket ID: %s", socket_id.hex())
            self.socket_ids.append(socket_id)
        return socket_id

    def _disconnect_remote_device(self, msg: dict):
        ip_key = msg['address'] + ':' + msg['port']
        if ip_key in self.tcp_ip_2_id_dict:
            self.tcpsocket.send(self.tcp_ip_2_id_dict[ip_key], zmq.SNDMORE)
            self.tcpsocket.send(b'', zmq.NOBLOCK)
            del self.tcp_ip_2_id_dict[ip_key]

    def _send_remote_device(self, r_device_id, msg):
        if r_device_id in self.remote_device_id_msg_dict:
            remote = self.remote_device_id_msg_dict[r_device_id]
            ip_address = remote['address']
            port = remote['port']
            url = f"tcp://{ip_address}:{port}"
            self.tcpsocket.connect(url)
            socket_id = self.tcpsocket.getsockopt(zmq.IDENTITY)
            self.tcpsocket.send(socket_id, zmq.SNDMORE)
            self.tcpsocket.send(msg)
            self.tcpsocket.send(socket_id, zmq.SNDMORE)
            self.tcpsocket.send(b'', zmq.NOBLOCK)

    def _service_zconf_message(self, rx_zconf_pull):
        message = rx_zconf_pull.recv()
        msg = json.loads(message)
        should_connect = False
        if msg['objectType'] in ['zeroConfAdd', 'zeroConfUpdate']:
            device_list = msg['deviceID'].split(',')
            for device_id in device_list:
                if len(device_id) > 0 and (device_id not in self.local_device_id_list) and (device_id not in self.remote_device_id_msg_dict):
                    self.remote_device_id_msg_dict[device_id] = msg
                    added = True
                    if device_id in self._device_id_action_station_list:
                        should_connect = True
                else:
                    added = False
            if added:
                self.remote_connection_dict[msg['connectionID']] = device_list
                self.remote_device_con_dict[msg['connectionID']] = msg
                added = False
                if should_connect:
                    self._connect_remote_device(msg)
                    should_connect = False
        elif msg['objectType'] == 'zeroConfDisconnect':
            try:
                for device_id in self.remote_connection_dict[msg['connectionID']]:
                    if device_id in self.remote_device_id_msg_dict:
                        self._disconnect_remote_device(self.remote_device_id_msg_dict[device_id])
                        del self.remote_device_id_msg_dict[device_id]
            except KeyError:
                return
            del self.remote_connection_dict[msg['connectionID']]
            del self.remote_device_con_dict[msg['connectionID']]

    def _add_device_rx(self, msg_dict):
        """Connect to another device"""
        device_id = msg_dict["deviceID"]
        logging.debug("TCP DEVICE CONNECT: %s", device_id)
        if device_id not in self._device_id_action_station_list:
            self._device_id_action_station_list.append(device_id)
        if device_id in self.remote_device_id_msg_dict:
            self._connect_remote_device(self.remote_device_id_msg_dict[device_id])
        logging.debug("device_ids to connect too: %s", self._device_id_action_station_list)

    def _del_device_rx(self, msg_dict):
        device_id = msg_dict["deviceID"]
        if device_id in self._device_id_action_station_list:
            logging.debug("TCP DEVICE_DISCONNECT: %s", device_id)
            if device_id in self.remote_device_id_msg_dict:
                self._disconnect_remote_device(self.remote_device_id_msg_dict[device_id])
            self._device_id_action_station_list.remove(device_id)
        logging.debug("device_ids to connect too: %s", self._device_id_action_station_list)

    def _tcp_command(self, msg_dict: dict):
        logging.debug("TCP CMD: %s", msg_dict)
        if msg_dict['msgType'] == 'connect':
            self._add_device_rx(msg_dict)
        if msg_dict['msgType'] == 'disconnect':
            self._del_device_rx(msg_dict)

    def _service_device_messaging(self):

        def _zmq_tcp_send(tcp_id, data: bytearray):
            logging.debug("TCP TX:\n%s", data.decode().rstrip())
            try:
                self.tcpsocket.send(tcp_id, zmq.SNDMORE)
                self.tcpsocket.send(data, zmq.NOBLOCK)
            except zmq.error.ZMQError as zmq_e:
                logging.debug("Sending TX Error: %s", zmq_e)
                #  self.socket_ids.remove(tcp_id)
            except OSError as exc:
                logging.debug("Socket assignment error: %s", exc)

        try:
            [msg_to, data] = self.rx_zmq_sub.recv_multipart()
        except ValueError:
            logging.debug("TCP value error")
            return
        if not data:
            logging.debug("TCP no data error")
            return
        if msg_to == b'ALL':
            for tcp_id in self.socket_ids:
                _zmq_tcp_send(tcp_id, data)
        elif msg_to == b'COMMAND':
            self._tcp_command(json.loads(data))
            return
        else:
            dest = msg_to.split(b':')[-1]
            if dest in self.socket_ids:
                _zmq_tcp_send(dest, data)
            if dest in self.remote_device_id_msg_dict:
                self._send_remote_device(dest, data)

    def _service_tcp_messages(self, tx_zmq_pub):
        tcp_id = self.tcpsocket.recv()
        message = self.tcpsocket.recv()
        if tcp_id not in self.socket_ids and tcp_id not in self.tcp_id_2_ip_dict:
            logging.debug("Added Socket ID: %s", tcp_id.hex())
            self.socket_ids.append(tcp_id)
        if message:
            logging.debug("TCP RX: %s\n%s", tcp_id.hex(), message.decode().rstrip())
            msg_from = self.b_zmq_connection_uuid + b":" + tcp_id
            tx_zmq_pub.send_multipart([message, msg_from])
        else:
            if tcp_id in self.socket_ids:
                logging.debug("Removed Socket ID: %s", tcp_id.hex())
                self.socket_ids.remove(tcp_id)

    def run(self):
        self.tcpsocket = self.context.socket(zmq.STREAM)

        self.rx_zmq_sub = self.context.socket(zmq.SUB)
        # Subscribe on ALL, COMMAND, and my zmq_connection_uuid
        self.rx_zmq_sub.setsockopt_string(zmq.SUBSCRIBE, "ALL")
        self.rx_zmq_sub.setsockopt_string(zmq.SUBSCRIBE, "COMMAND")
        self.rx_zmq_sub.setsockopt_string(zmq.SUBSCRIBE, "TCP")
        self.rx_zmq_sub.setsockopt_string(zmq.SUBSCRIBE, self.zmq_connection_uuid)

        self.tcpsocket.bind(self.ext_url)
        self.tcpsocket.set(zmq.SNDTIMEO, 5)

        rx_zconf_pull = self.context.socket(zmq.PULL)
        rx_zconf_pull.bind("inproc://zconf")

        poller = zmq.Poller()
        poller.register(self.tcpsocket, zmq.POLLIN)
        poller.register(self.rx_zmq_sub, zmq.POLLIN)
        poller.register(rx_zconf_pull, zmq.POLLIN)

        self._send_announce()

        while self.running:
            try:
                socks = dict(poller.poll(50))
            except zmq.error.ContextTerminated:
                break
            if self.tcpsocket in socks:
                self._service_tcp_messages(self.tx_zmq_pub)
            if self.rx_zmq_sub in socks:
                self._service_device_messaging()
            if rx_zconf_pull in socks:
                self._service_zconf_message(rx_zconf_pull)

        for tcp_id in self.socket_ids:
            self.tcpsocket.send(tcp_id, zmq.SNDMORE)
            self.tcpsocket.send(b'', zmq.NOBLOCK)

        self.tcpsocket.close()
        self.tx_zmq_pub.close()
        self.rx_zmq_sub.close()
        rx_zconf_pull.close()



"""
MIT License

Copyright (c) 2020 DashIO-Connect

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from ..constants import BAD_CHARS
from .control import Control, ControlPosition, ControlConfig, _get_color, _get_icon, _get_title_position
from .enums import ButtonState, Color, Icon, TitlePosition


class ButtonConfig(ControlConfig):
    """ButtonGroupConfig"""
    def __init__(
        self,
        control_id: str,
        title: str,
        title_position: TitlePosition,
        button_enabled: bool,
        icon_name: Icon,
        on_color: Color,
        off_color: Color,
        text: str,
        control_position: ControlPosition
    ) -> None:
        super().__init__(control_id, title, control_position, title_position)
        self.cfg["text"] = text.translate(BAD_CHARS)
        self.cfg["iconName"] = icon_name.value
        self.cfg["buttonEnabled"] = button_enabled
        self.cfg["onColor"] = str(on_color.value)
        self.cfg["offColor"] = str(off_color.value)

    @classmethod
    def from_dict(cls, cfg_dict: dict):
        """Instantiates ButtonGroupConfig from cfg dictionary

        Parameters
        ----------
        cfg_dict : dict
            A dictionary usually loaded from a config json from IoTDashboard App

        Returns
        -------
        ButtonGroupConfig
        """
        tmp_cls = cls(
            cfg_dict["controlID"],
            cfg_dict["title"],
            _get_title_position(cfg_dict["titlePosition"]),
            cfg_dict["buttonEnabled"],
            _get_icon(cfg_dict["iconName"]),
            _get_color(cfg_dict["onColor"]),
            _get_color(cfg_dict["offColor"]),
            cfg_dict["text"],
            ControlPosition(cfg_dict["xPositionRatio"], cfg_dict["yPositionRatio"], cfg_dict["widthRatio"], cfg_dict["heightRatio"])
        )
        tmp_cls.parent_id = cfg_dict["parentID"]
        return tmp_cls


class Button(Control):
    """A Button control.

    Attributes
    ----------
    control_id : str
        An unique control identity string. The control identity string must be a unique string for each control per device
    title: str
        A short title for the button group
    text : str
        The text that appears on the ButtonGroup
    title_position : TitlePosition
        Can be TitlePosition.BOTTOM, TitlePosition.TOP, TitlePosition.OFF
    button_enabled : boolean
        True allows the app to send button events. False disables button pushes
    icon_name : Icon
        Set the icon for the button
    off_color : Color
        Set the off color
    on_color : Color
        Set the on color
    control_position : ControlPosition
        Set the size and position of the button on a DeviceView.

    Methods
    -------
    send_button(btn_state, btn_icon, text) :
        Updates the button state, button icon and text.

    from_cfg_dict(cfg_dict: dict) :
        Instatiates Button from cfg dictionary
    """

    def toggle_btn(self):
        """
        Toggles the current ButtonState
        """
        if self.btn_state == ButtonState.FLASH:
            return
        if self.btn_state == ButtonState.OFF:
            self.btn_state = ButtonState.ON
        else:
            self.btn_state = ButtonState.OFF

    def __init__(
        self,
        control_id,
        title="A Button",
        title_position=TitlePosition.BOTTOM,
        button_enabled=True,
        icon_name=Icon.NONE,
        on_color=Color.BLUE,
        off_color=Color.RED,
        text="",
        control_position: ControlPosition = None
    ):
        super().__init__("BTTN", control_id)
        self._cfg_columnar.append(
            ButtonConfig(
                control_id,
                title,
                title_position,
                button_enabled,
                icon_name,
                on_color,
                off_color,
                text,
                control_position
            )
        )
        self._btn_state = ButtonState.OFF
        self._text = text.translate(BAD_CHARS)
        self._icon_name = icon_name

    @classmethod
    def from_cfg_dict(cls, cfg_dict: dict):
        """Instatiates Button from cfg dictionary

        Parameters
        ----------
        cfg_dict : dict
            A dictionary usually loaded from a config json from IoTDashboard App

        Returns
        -------
        Button
        """
        tmp_cls = cls(
            cfg_dict["controlID"],
            cfg_dict["title"],
            _get_title_position(cfg_dict["titlePosition"]),
            cfg_dict["buttonEnabled"],
            _get_icon(cfg_dict["iconName"]),
            _get_color(cfg_dict["onColor"]),
            _get_color(cfg_dict["offColor"]),
            cfg_dict["text"],
            ControlPosition(cfg_dict["xPositionRatio"], cfg_dict["yPositionRatio"], cfg_dict["widthRatio"], cfg_dict["heightRatio"])
        )
        tmp_cls.parent_id = cfg_dict["parentID"]
        return tmp_cls

    def get_state(self):
        """get_state is called by iotdashboard

        Returns
        -------
        str
            The controls state
        """
        text = self._text
        if (not text) and (self.icon_name == Icon.NONE):
            return self._control_hdr_str + f"{self._btn_state.value}\n"
        if (not text) and self.icon_name != Icon.NONE:
            return self._control_hdr_str + f"{self._btn_state.value}\t{self._icon_name.value}\n"
        return self._control_hdr_str + f"{self._btn_state.value}\t{self._icon_name.value}\t{text}\n"

    @property
    def icon_name(self) -> Icon:
        """Returns the Icon the Button is set too

        Returns
        -------
        Icon
            The Icon the button is currently set to
        """
        return self._icon_name

    @icon_name.setter
    def icon_name(self, val: Icon):
        self._icon_name = val
        self.state_str = self._control_hdr_str + f"{self._btn_state.value}\t{val.value}\n"

    @property
    def text(self) -> str:
        """Return the Buttons current text string

        Returns
        -------
        str
            The buttons text.
        """

        return self._text

    @text.setter
    def text(self, val: str):
        self._text = val.translate(BAD_CHARS)
        self.state_str = self._control_hdr_str + f"{self._btn_state.value}\t{self._icon_name.value}\t{self._text}\n"

    @property
    def btn_state(self) -> ButtonState:
        """
        Returns the state of the button.
        """
        return self._btn_state

    @btn_state.setter
    def btn_state(self, val: ButtonState):
        self._btn_state = val
        self.state_str = self._control_hdr_str + f"{val.value}\n"

    def send_button(self, btn_state: ButtonState, btn_icon: Icon, text: str):
        """Sends the button state to **DashIO** app.

        Parameters
        ----------
        btn_state : ButtonState
            State of the button
        btn_icon : Icon
            Icon to send
        text : str
            Text to send.
        """
        self._btn_state = btn_state
        self._icon_name = btn_icon
        self._text = text.translate(BAD_CHARS)
        self.state_str = self._control_hdr_str + f"{self._btn_state.value}\t{self._icon_name.value}\t{self._text}\n"



"""
MIT License

Copyright (c) 2020 DashIO-Connect

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import logging
import threading

import zmq

from ..constants import CONNECTION_PUB_URL, TASK_PULL
from .action_station_service_config import (
    ActionServiceCFG,
    IntParameterSpec,
    SelectorParameterSpec
)


# This defines the provisioning for the TIMER
def make_timer_config(num_timers):
    """Make a timer config"""
    provisioning_list = [
        SelectorParameterSpec("Timer Type", ["Repeat", "OneShot"], "Repeat"),
        IntParameterSpec("Timeout", 100, 600000, "ms", 1000)
    ]
    parameter_in_list = []
    #  parameter_out_list = []

    timer_cfg = ActionServiceCFG(
        "TMR",
        "Timer",
        "A timer control.",
        "TMR",
        num_timers,
        True,
        True,
        provisioning_list,
        parameter_in_list
        #  parameter_out_list
    )
    return timer_cfg.__json__()


class RepeatTimer(threading.Timer):
    """The timer"""
    def run(self):
        while not self.finished.wait(self.interval):
            self.function(*self.args, **self.kwargs)


# The app returns the edited provisioning with enough info to instantiate this class:
# With controlID, timer_type(from provisioning), timeout(from provisioning)
# THe device knows the rest (device_id, push_url, context)

# parameter_in_list would be used by the TIMER to parse the incoming message.
# paramiter_out_list is how to format the output message
class TimerService(threading.Thread):
    """Timer Class"""

    def timer_message(self):
        """Send the message"""
        self.send_message(out_message=self.control_msg + "\n")

    def send_message(self, out_message=""):
        """Send the message"""
        self.task_sender.send_multipart([b"ALL", out_message.encode('utf-8')])

    def close(self):
        """Close the thread"""
        self.running = False

    def __init__(self, device_id: str, action_station_id: str, control_config_dict: dict, context: zmq.Context) -> None:
        threading.Thread.__init__(self, daemon=True)

        self.context = context
        self.running = True
        self.timer_type = None

        self.control_id = control_config_dict['controlID']
        self.name = control_config_dict['name']
        self.control_type = control_config_dict['objectType']
        provision_list = control_config_dict['provisioning']

        self.sub_url = CONNECTION_PUB_URL.format(id=action_station_id)

        self.push_url = TASK_PULL.format(id=action_station_id)
        self.task_sender = self.context.socket(zmq.PUSH)
        self.task_sender.connect(self.push_url)

        self.timer_time = int(provision_list[1]['value']) / 1000.0
        self.timer_type = provision_list[0]['value']

        self.control_msg = f"\t{device_id}\t{self.control_type}\t{self.control_id}"

        logging.debug("Init Class: %s, %s", self.control_type, self.name)

        if self.timer_type == 'Repeat':
            self.timer_type = RepeatTimer(self.timer_time, self.timer_message)
            self.timer_type.start()
        self.start()

    def run(self):
        receiver = self.context.socket(zmq.SUB)
        receiver.connect(self.sub_url)
        receiver.setsockopt_string(zmq.SUBSCRIBE, self.control_msg)
        poller = zmq.Poller()
        poller.register(receiver, zmq.POLLIN)

        while self.running:
            try:
                socks = dict(poller.poll(15))
            except zmq.error.ContextTerminated:
                break
            if receiver in socks:
                message, _ = receiver.recv()
                if message:
                    logging.debug("%s\t%s RX:\n%s", self.control_type, self.control_id, message.decode())
                    if self.timer_type == 'OneShot':
                        self.timer_type = threading.Timer(self.timer_time, self.timer_message)
                        self.timer_type.start()
        self.timer_type.cancel()
        self.task_sender.close()
        receiver.close()
